else(8)         Jump: 86        Pop: 1                          { align1 switch };
else(16)        Jump: 86        Pop: 1                          { align1 switch };
else(8)         Jump: 14        Pop: 1                          { align16 switch };
