#ifndef NV50_COMPUTE_XML
#define NV50_COMPUTE_XML

/* Autogenerated file, DO NOT EDIT manually!

This file was generated by the rules-ng-ng headergen tool in this git repository:
http://github.com/envytools/envytools/
git clone https://github.com/envytools/envytools.git

The rules-ng-ng source files this header was generated from are:
- rnndb/graph/g80_compute.xml (  14027 bytes, from 2015-02-14 02:01:36)
- rnndb/copyright.xml         (   6456 bytes, from 2015-02-14 02:01:36)
- rnndb/nvchipsets.xml        (   2833 bytes, from 2015-04-28 16:28:33)
- rnndb/fifo/nv_object.xml    (  15390 bytes, from 2015-04-22 20:36:09)
- rnndb/g80_defs.xml          (  18210 bytes, from 2015-10-19 20:49:59)

Copyright (C) 2006-2015 by the following authors:
- Artur Huillet <arthur.huillet@free.fr> (ahuillet)
- Ben Skeggs (darktama, darktama_)
- B. R. <koala_br@users.sourceforge.net> (koala_br)
- Carlos Martin <carlosmn@users.sf.net> (carlosmn)
- Christoph Bumiller <e0425955@student.tuwien.ac.at> (calim, chrisbmr)
- Dawid Gajownik <gajownik@users.sf.net> (gajownik)
- Dmitry Baryshkov
- Dmitry Eremin-Solenikov <lumag@users.sf.net> (lumag)
- EdB <edb_@users.sf.net> (edb_)
- Erik Waling <erikwailing@users.sf.net> (erikwaling)
- Francisco Jerez <currojerez@riseup.net> (curro)
- Ilia Mirkin <imirkin@alum.mit.edu> (imirkin)
- jb17bsome <jb17bsome@bellsouth.net> (jb17bsome)
- Jeremy Kolb <kjeremy@users.sf.net> (kjeremy)
- Laurent Carlier <lordheavym@gmail.com> (lordheavy)
- Luca Barbieri <luca@luca-barbieri.com> (lb, lb1)
- Maarten Maathuis <madman2003@gmail.com> (stillunknown)
- Marcin Kościelnicki <koriakin@0x04.net> (mwk, koriakin)
- Mark Carey <mark.carey@gmail.com> (careym)
- Matthieu Castet <matthieu.castet@parrot.com> (mat-c)
- nvidiaman <nvidiaman@users.sf.net> (nvidiaman)
- Patrice Mandin <patmandin@gmail.com> (pmandin, pmdata)
- Pekka Paalanen <pq@iki.fi> (pq, ppaalanen)
- Peter Popov <ironpeter@users.sf.net> (ironpeter)
- Richard Hughes <hughsient@users.sf.net> (hughsient)
- Rudi Cilibrasi <cilibrar@users.sf.net> (cilibrar)
- Serge Martin
- Simon Raffeiner
- Stephane Loeuillet <leroutier@users.sf.net> (leroutier)
- Stephane Marchesin <stephane.marchesin@gmail.com> (marcheu)
- sturmflut <sturmflut@users.sf.net> (sturmflut)
- Sylvain Munaut <tnt@246tNt.com>
- Victor Stinner <victor.stinner@haypocalc.com> (haypo)
- Wladmir van der Laan <laanwj@gmail.com> (miathan6)
- Younes Manton <younes.m@gmail.com> (ymanton)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/



#define NV50_COMPUTE_DMA_NOTIFY					0x00000180

#define NV50_COMPUTE_DMA_GLOBAL					0x000001a0

#define NV50_COMPUTE_DMA_QUERY					0x000001a4

#define NV50_COMPUTE_DMA_LOCAL					0x000001b8

#define NV50_COMPUTE_DMA_STACK					0x000001bc

#define NV50_COMPUTE_DMA_CODE_CB					0x000001c0

#define NV50_COMPUTE_DMA_TSC					0x000001c4

#define NV50_COMPUTE_DMA_TIC					0x000001c8

#define NV50_COMPUTE_DMA_TEXTURE					0x000001cc

#define NV50_COMPUTE_UNK0200					0x00000200
#define NV50_COMPUTE_UNK0200_UNK1__MASK				0x0000ffff
#define NV50_COMPUTE_UNK0200_UNK1__SHIFT				0
#define NV50_COMPUTE_UNK0200_UNK2__MASK				0x00ff0000
#define NV50_COMPUTE_UNK0200_UNK2__SHIFT				16

#define NV50_COMPUTE_UNK0204					0x00000204

#define NV50_COMPUTE_UNK0208					0x00000208

#define NV50_COMPUTE_UNK020C					0x0000020c

#define NV50_COMPUTE_CP_ADDRESS_HIGH				0x00000210

#define NV50_COMPUTE_CP_ADDRESS_LOW				0x00000214

#define NV50_COMPUTE_STACK_ADDRESS_HIGH				0x00000218

#define NV50_COMPUTE_STACK_ADDRESS_LOW				0x0000021c

#define NV50_COMPUTE_STACK_SIZE_LOG				0x00000220

#define NV50_COMPUTE_CALL_LIMIT_LOG				0x00000224

#define NV50_COMPUTE_UNK0228					0x00000228
#define NV50_COMPUTE_UNK0228_UNK0				0x00000001
#define NV50_COMPUTE_UNK0228_UNK4__MASK				0x00000ff0
#define NV50_COMPUTE_UNK0228_UNK4__SHIFT				4
#define NV50_COMPUTE_UNK0228_UNK12__MASK				0x000ff000
#define NV50_COMPUTE_UNK0228_UNK12__SHIFT			12

#define NV50_COMPUTE_TSC_ADDRESS_HIGH				0x0000022c

#define NV50_COMPUTE_TSC_ADDRESS_LOW				0x00000230
#define NV50_COMPUTE_TSC_ADDRESS_LOW__ALIGN			0x00000020

#define NV50_COMPUTE_TSC_LIMIT					0x00000234
#define NV50_COMPUTE_TSC_LIMIT__MAX				0x00001fff

#define NV50_COMPUTE_CB_ADDR					0x00000238
#define NV50_COMPUTE_CB_ADDR_ID__MASK				0x003fff00
#define NV50_COMPUTE_CB_ADDR_ID__SHIFT				8
#define NV50_COMPUTE_CB_ADDR_BUFFER__MASK			0x0000007f
#define NV50_COMPUTE_CB_ADDR_BUFFER__SHIFT			0

#define NV50_COMPUTE_CB_DATA(i0)				       (0x0000023c + 0x4*(i0))
#define NV50_COMPUTE_CB_DATA__ESIZE				0x00000004
#define NV50_COMPUTE_CB_DATA__LEN				0x00000010

#define NV50_COMPUTE_TSC_FLUSH					0x0000027c
#define NV50_COMPUTE_TSC_FLUSH_SPECIFIC				0x00000001
#define NV50_COMPUTE_TSC_FLUSH_ENTRY__MASK			0x03fffff0
#define NV50_COMPUTE_TSC_FLUSH_ENTRY__SHIFT			4

#define NV50_COMPUTE_TIC_FLUSH					0x00000280
#define NV50_COMPUTE_TIC_FLUSH_SPECIFIC				0x00000001
#define NV50_COMPUTE_TIC_FLUSH_ENTRY__MASK			0x03fffff0
#define NV50_COMPUTE_TIC_FLUSH_ENTRY__SHIFT			4

#define NV50_COMPUTE_DELAY1					0x00000284

#define NV50_COMPUTE_WATCHDOG_TIMER				0x00000288

#define NV50_COMPUTE_DELAY2					0x0000028c

#define NV50_COMPUTE_UNK0290					0x00000290

#define NV50_COMPUTE_LOCAL_ADDRESS_HIGH				0x00000294

#define NV50_COMPUTE_LOCAL_ADDRESS_LOW				0x00000298
#define NV50_COMPUTE_LOCAL_ADDRESS_LOW__ALIGN			0x00000100

#define NV50_COMPUTE_LOCAL_SIZE_LOG				0x0000029c

#define NV50_COMPUTE_UNK02A0					0x000002a0

#define NV50_COMPUTE_CB_DEF_ADDRESS_HIGH				0x000002a4

#define NV50_COMPUTE_CB_DEF_ADDRESS_LOW				0x000002a8

#define NV50_COMPUTE_CB_DEF_SET					0x000002ac
#define NV50_COMPUTE_CB_DEF_SET_SIZE__MASK			0x0000ffff
#define NV50_COMPUTE_CB_DEF_SET_SIZE__SHIFT			0
#define NV50_COMPUTE_CB_DEF_SET_BUFFER__MASK			0x007f0000
#define NV50_COMPUTE_CB_DEF_SET_BUFFER__SHIFT			16

#define NV50_COMPUTE_UNK02B0					0x000002b0

#define NV50_COMPUTE_BLOCK_ALLOC					0x000002b4
#define NV50_COMPUTE_BLOCK_ALLOC_THREADS__MASK			0x0000ffff
#define NV50_COMPUTE_BLOCK_ALLOC_THREADS__SHIFT			0
#define NV50_COMPUTE_BLOCK_ALLOC_BARRIERS__MASK			0x00ff0000
#define NV50_COMPUTE_BLOCK_ALLOC_BARRIERS__SHIFT			16

#define NV50_COMPUTE_LANES32_ENABLE				0x000002b8

#define NV50_COMPUTE_UNK02BC					0x000002bc
#define NV50_COMPUTE_UNK02BC_UNK1__MASK				0x00000007
#define NV50_COMPUTE_UNK02BC_UNK1__SHIFT				0
#define NV50_COMPUTE_UNK02BC_UNK2__MASK				0x00000070
#define NV50_COMPUTE_UNK02BC_UNK2__SHIFT				4

#define NV50_COMPUTE_CP_REG_ALLOC_TEMP				0x000002c0

#define NV50_COMPUTE_TIC_ADDRESS_HIGH				0x000002c4

#define NV50_COMPUTE_TIC_ADDRESS_LOW				0x000002c8

#define NV50_COMPUTE_TIC_LIMIT					0x000002cc

#define NV50_COMPUTE_MP_PM_SET(i0)			       (0x000002d0 + 0x4*(i0))
#define NV50_COMPUTE_MP_PM_SET__ESIZE				0x00000004
#define NV50_COMPUTE_MP_PM_SET__LEN				0x00000004

#define NV50_COMPUTE_MP_PM_CONTROL(i0)			       (0x000002e0 + 0x4*(i0))
#define NV50_COMPUTE_MP_PM_CONTROL__ESIZE			0x00000004
#define NV50_COMPUTE_MP_PM_CONTROL__LEN				0x00000004
#define NV50_COMPUTE_MP_PM_CONTROL_MODE__MASK			0x00000001
#define NV50_COMPUTE_MP_PM_CONTROL_MODE__SHIFT			0
#define NV50_COMPUTE_MP_PM_CONTROL_MODE_LOGOP			0x00000000
#define NV50_COMPUTE_MP_PM_CONTROL_MODE_LOGOP_PULSE		0x00000001
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT__MASK			0x00000070
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT__SHIFT			4
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK0			0x00000000
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK1			0x00000010
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK2			0x00000020
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK3			0x00000030
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK4			0x00000040
#define NV50_COMPUTE_MP_PM_CONTROL_UNIT_UNK5			0x00000050
#define NV50_COMPUTE_MP_PM_CONTROL_FUNC__MASK			0x00ffff00
#define NV50_COMPUTE_MP_PM_CONTROL_FUNC__SHIFT			8
#define NV50_COMPUTE_MP_PM_CONTROL_SIG__MASK			0xff000000
#define NV50_COMPUTE_MP_PM_CONTROL_SIG__SHIFT			24

#define NV50_COMPUTE_MP_PM_OVERFLOW_TRAP_ENABLE			0x000002f0
#define NV50_COMPUTE_MP_PM_OVERFLOW_TRAP_ENABLE_0		0x00000001
#define NV50_COMPUTE_MP_PM_OVERFLOW_TRAP_ENABLE_1		0x00000002
#define NV50_COMPUTE_MP_PM_OVERFLOW_TRAP_ENABLE_2		0x00000004
#define NV50_COMPUTE_MP_PM_OVERFLOW_TRAP_ENABLE_3		0x00000008

#define NV50_COMPUTE_UNK02F4					0x000002f4

#define NV50_COMPUTE_BLOCKDIM_LATCH				0x000002f8

#define NV50_COMPUTE_LOCAL_WARPS_LOG_ALLOC			0x000002fc

#define NV50_COMPUTE_LOCAL_WARPS_NO_CLAMP			0x00000300

#define NV50_COMPUTE_STACK_WARPS_LOG_ALLOC			0x00000304

#define NV50_COMPUTE_STACK_WARPS_NO_CLAMP			0x00000308

#define NV50_COMPUTE_UNK030C					0x0000030c

#define NV50_COMPUTE_QUERY_ADDRESS_HIGH				0x00000310

#define NV50_COMPUTE_QUERY_ADDRESS_LOW				0x00000314

#define NV50_COMPUTE_QUERY_SEQUENCE				0x00000318

#define NV50_COMPUTE_QUERY_GET					0x0000031c
#define NV50_COMPUTE_QUERY_GET_INTR				0x00000200
#define NV50_COMPUTE_QUERY_GET_SHORT				0x00008000

#define NV50_COMPUTE_COND_ADDRESS_HIGH				0x00000320

#define NV50_COMPUTE_COND_ADDRESS_LOW				0x00000324

#define NV50_COMPUTE_COND_MODE					0x00000328
#define NV50_COMPUTE_COND_MODE_NEVER				0x00000000
#define NV50_COMPUTE_COND_MODE_ALWAYS				0x00000001
#define NV50_COMPUTE_COND_MODE_RES_NON_ZERO			0x00000002
#define NV50_COMPUTE_COND_MODE_EQUAL				0x00000003
#define NV50_COMPUTE_COND_MODE_NOT_EQUAL				0x00000004

#define NV50_COMPUTE_UNK032C					0x0000032c

#define NV50_COMPUTE_UNK0330					0x00000330

#define NV50_COMPUTE_UNK0334(i0)				       (0x00000334 + 0x4*(i0))
#define NV50_COMPUTE_UNK0334__ESIZE				0x00000004
#define NV50_COMPUTE_UNK0334__LEN				0x00000003

#define NV50_COMPUTE_UNK0340(i0)				       (0x00000340 + 0x4*(i0))
#define NV50_COMPUTE_UNK0340__ESIZE				0x00000004
#define NV50_COMPUTE_UNK0340__LEN				0x00000002

#define NV50_COMPUTE_UNK0348(i0)				       (0x00000348 + 0x4*(i0))
#define NV50_COMPUTE_UNK0348__ESIZE				0x00000004
#define NV50_COMPUTE_UNK0348__LEN				0x00000002

#define NV50_COMPUTE_UNK0350(i0)				       (0x00000350 + 0x4*(i0))
#define NV50_COMPUTE_UNK0350__ESIZE				0x00000004
#define NV50_COMPUTE_UNK0350__LEN				0x00000002

#define NV50_COMPUTE_UNK0358					0x00000358

#define NV50_COMPUTE_UNK035C					0x0000035c

#define NV50_COMPUTE_UNK0360					0x00000360
#define NV50_COMPUTE_UNK0360_UNK0__MASK				0x000000f0
#define NV50_COMPUTE_UNK0360_UNK0__SHIFT				4
#define NV50_COMPUTE_UNK0360_UNK1__MASK				0x00000f00
#define NV50_COMPUTE_UNK0360_UNK1__SHIFT				8

#define NV50_COMPUTE_UNK0364					0x00000364

#define NV50_COMPUTE_LAUNCH					0x00000368

#define NV50_COMPUTE_UNK036C					0x0000036c

#define NV50_COMPUTE_UNK0370					0x00000370

#define NV50_COMPUTE_USER_PARAM_COUNT				0x00000374
#define NV50_COMPUTE_USER_PARAM_COUNT_UNK0__MASK			0x000000ff
#define NV50_COMPUTE_USER_PARAM_COUNT_UNK0__SHIFT		0
#define NV50_COMPUTE_USER_PARAM_COUNT_COUNT__MASK		0x0000ff00
#define NV50_COMPUTE_USER_PARAM_COUNT_COUNT__SHIFT		8
#define NV50_COMPUTE_USER_PARAM_COUNT_COUNT__MAX			0x00000040

#define NV50_COMPUTE_LINKED_TSC					0x00000378

#define NV50_COMPUTE_UNK037C					0x0000037c
#define NV50_COMPUTE_UNK037C_ALWAYS_DERIV			0x00000001
#define NV50_COMPUTE_UNK037C_UNK16				0x00010000

#define NV50_COMPUTE_CODE_CB_FLUSH				0x00000380

#define NV50_COMPUTE_UNK0384					0x00000384

#define NV50_COMPUTE_GRIDID					0x00000388

#define NV50_COMPUTE_UNK038C(i0)				       (0x0000038c + 0x4*(i0))
#define NV50_COMPUTE_UNK038C__ESIZE				0x00000004
#define NV50_COMPUTE_UNK038C__LEN				0x00000003

#define NV50_COMPUTE_WRCACHE_FLUSH				0x00000398

#define NV50_COMPUTE_UNK039C(i0)				       (0x0000039c + 0x4*(i0))
#define NV50_COMPUTE_UNK039C__ESIZE				0x00000004
#define NV50_COMPUTE_UNK039C__LEN				0x00000002

#define NV50_COMPUTE_GRIDDIM					0x000003a4
#define NV50_COMPUTE_GRIDDIM_X__MASK				0x0000ffff
#define NV50_COMPUTE_GRIDDIM_X__SHIFT				0
#define NV50_COMPUTE_GRIDDIM_Y__MASK				0xffff0000
#define NV50_COMPUTE_GRIDDIM_Y__SHIFT				16

#define NV50_COMPUTE_SHARED_SIZE					0x000003a8
#define NV50_COMPUTE_SHARED_SIZE__MAX				0x00004000
#define NV50_COMPUTE_SHARED_SIZE__ALIGN				0x00000040

#define NV50_COMPUTE_BLOCKDIM_XY					0x000003ac
#define NV50_COMPUTE_BLOCKDIM_XY_X__MASK				0x0000ffff
#define NV50_COMPUTE_BLOCKDIM_XY_X__SHIFT			0
#define NV50_COMPUTE_BLOCKDIM_XY_Y__MASK				0xffff0000
#define NV50_COMPUTE_BLOCKDIM_XY_Y__SHIFT			16

#define NV50_COMPUTE_BLOCKDIM_Z					0x000003b0
#define NV50_COMPUTE_BLOCKDIM_Z__MIN				0x00000001
#define NV50_COMPUTE_BLOCKDIM_Z__MAX				0x00000040

#define NV50_COMPUTE_CP_START_ID					0x000003b4

#define NV50_COMPUTE_REG_MODE					0x000003b8
#define NV50_COMPUTE_REG_MODE_PACKED				0x00000001
#define NV50_COMPUTE_REG_MODE_STRIPED				0x00000002

#define NV50_COMPUTE_TEX_LIMITS					0x000003bc
#define NV50_COMPUTE_TEX_LIMITS_SAMPLERS_LOG2__MASK		0x0000000f
#define NV50_COMPUTE_TEX_LIMITS_SAMPLERS_LOG2__SHIFT		0
#define NV50_COMPUTE_TEX_LIMITS_SAMPLERS_LOG2__MIN		0x00000000
#define NV50_COMPUTE_TEX_LIMITS_SAMPLERS_LOG2__MAX		0x00000004
#define NV50_COMPUTE_TEX_LIMITS_TEXTURES_LOG2__MASK		0x000000f0
#define NV50_COMPUTE_TEX_LIMITS_TEXTURES_LOG2__SHIFT		4
#define NV50_COMPUTE_TEX_LIMITS_TEXTURES_LOG2__MIN		0x00000000
#define NV50_COMPUTE_TEX_LIMITS_TEXTURES_LOG2__MAX		0x00000007

#define NV50_COMPUTE_BIND_TSC					0x000003c0
#define NV50_COMPUTE_BIND_TSC_VALID				0x00000001
#define NV50_COMPUTE_BIND_TSC_SAMPLER__MASK			0x000000f0
#define NV50_COMPUTE_BIND_TSC_SAMPLER__SHIFT			4
#define NV50_COMPUTE_BIND_TSC_TSC__MASK				0x001ff000
#define NV50_COMPUTE_BIND_TSC_TSC__SHIFT				12

#define NV50_COMPUTE_BIND_TIC					0x000003c4
#define NV50_COMPUTE_BIND_TIC_VALID				0x00000001
#define NV50_COMPUTE_BIND_TIC_TEXTURE__MASK			0x000001fe
#define NV50_COMPUTE_BIND_TIC_TEXTURE__SHIFT			1
#define NV50_COMPUTE_BIND_TIC_TIC__MASK				0x7ffffe00
#define NV50_COMPUTE_BIND_TIC_TIC__SHIFT				9

#define NV50_COMPUTE_SET_PROGRAM_CB				0x000003c8
#define NV50_COMPUTE_SET_PROGRAM_CB_INDEX__MASK			0x00000f00
#define NV50_COMPUTE_SET_PROGRAM_CB_INDEX__SHIFT			8
#define NV50_COMPUTE_SET_PROGRAM_CB_BUFFER__MASK			0x0007f000
#define NV50_COMPUTE_SET_PROGRAM_CB_BUFFER__SHIFT		12
#define NV50_COMPUTE_SET_PROGRAM_CB_VALID			0x000000ff

#define NV50_COMPUTE_UNK03CC					0x000003cc

#define NV50_COMPUTE_TEX_CACHE_CTL				0x000003d0
#define NV50_COMPUTE_TEX_CACHE_CTL_UNK1__MASK			0x00000030
#define NV50_COMPUTE_TEX_CACHE_CTL_UNK1__SHIFT			4

#define NV50_COMPUTE_UNK03D4					0x000003d4

#define NV50_COMPUTE_UNK03D8					0x000003d8

#define NV50_COMPUTE_UNK03DC					0x000003dc

#define NV50_COMPUTE_UNK03E0					0x000003e0

#define NV50_COMPUTE_UNK03E4					0x000003e4

#define NVA3_COMPUTE_TEX_MISC					0x000003e8
#define NVA3_COMPUTE_TEX_MISC_UNK1				0x00000001
#define NVA3_COMPUTE_TEX_MISC_SEAMLESS_CUBE_MAP		0x00000002

#define NV50_COMPUTE_GLOBAL(i0)				       (0x00000400 + 0x20*(i0))
#define NV50_COMPUTE_GLOBAL__ESIZE				0x00000020
#define NV50_COMPUTE_GLOBAL__LEN					0x00000010

#define NV50_COMPUTE_GLOBAL_ADDRESS_HIGH(i0)		       (0x00000400 + 0x20*(i0))

#define NV50_COMPUTE_GLOBAL_ADDRESS_LOW(i0)		       (0x00000404 + 0x20*(i0))

#define NV50_COMPUTE_GLOBAL_PITCH(i0)			       (0x00000408 + 0x20*(i0))
#define NV50_COMPUTE_GLOBAL_PITCH__MAX				0x00800000
#define NV50_COMPUTE_GLOBAL_PITCH__ALIGN				0x00000100

#define NV50_COMPUTE_GLOBAL_LIMIT(i0)			       (0x0000040c + 0x20*(i0))

#define NV50_COMPUTE_GLOBAL_MODE(i0)			       (0x00000410 + 0x20*(i0))
#define NV50_COMPUTE_GLOBAL_MODE_LINEAR				0x00000001
#define NV50_COMPUTE_GLOBAL_MODE_UNK1__MASK			0x000000f0
#define NV50_COMPUTE_GLOBAL_MODE_UNK1__SHIFT			4
#define NV50_COMPUTE_GLOBAL_MODE_TILE_MODE__MASK			0x00000f00
#define NV50_COMPUTE_GLOBAL_MODE_TILE_MODE__SHIFT		8

#define NV50_COMPUTE_USER_PARAM(i0)			       (0x00000600 + 0x4*(i0))
#define NV50_COMPUTE_USER_PARAM__ESIZE				0x00000004
#define NV50_COMPUTE_USER_PARAM__LEN				0x00000040

#define NV50_COMPUTE_UNK0700(i0)				       (0x00000700 + 0x4*(i0))
#define NV50_COMPUTE_UNK0700__ESIZE				0x00000004
#define NV50_COMPUTE_UNK0700__LEN				0x00000010


#endif /* NV50_COMPUTE_XML */
