1. Number of dalmations: __LINE__ __FILE__ __LINE__
2. Nominal visual acuity: __LINE__ __FILE__ / __LINE__ __FILE__
3. Battle of Thermopylae, as film: __LINE__ __FILE__ __FILE__
4. HTTP code for Not Found: __LINE__ __FILE__ __LINE__
5. Hexadecimal for 20560: __LINE__ __FILE__ __LINE__ __FILE__
6: Zip code for Nortonville, KS: __LINE__ __LINE__ __FILE__ __LINE__ __FILE__
7. James Bond, as a number: __FILE__ __FILE__ __LINE__
