/*
 * Copyright 2023 Google LLC
 * SPDX-License-Identifier: MIT
 */

#include "gfxstream_vk_entrypoints.h"
#include "gfxstream_vk_private.h"

VkResult gfxstream_vk_wsi_init(struct gfxstream_vk_physical_device* physical_device) {
    return VK_SUCCESS;
}

void gfxstream_vk_wsi_finish(struct gfxstream_vk_physical_device* physical_device) {}
